#include <stdarg.h>
#include <stdio.h>

#include "lib/log.h"
#include "levdyn.h"
#include "things.h"
#include "updmap.h"
#include "game.h"
#include "gettable.h"
#include "linetype.h"

static int find_taggedsect(LevData *ld,int tag,int i) {
   while(i<ldnsectors(ld)) {
      if(ldsector(ld)[i].tag==tag) return i;
      i++;
   };
   return -1;
};

static int find_spot(LevData *ld,const LineType *lt,int tag) {
   int i;
   int sector=find_taggedsect(ld,tag,0); /* TODO: pick a random tagged sect */
   const ThingDyn *td=ldthingd(ld);
   logprintf(LOG_DEBUG,'M',"find_spot: tag=%d sector=%d spottype=%d",
	     tag,sector,(int)(lt->spottype));
   if(sector<0) return -1;
   for(i=0;i<ldnthings(ld);i++,td++)
      if(td->proto&&td->proto->id==lt->spottype&&td->sector==sector) {
	 logprintf(LOG_DEBUG,'M',"find_spot: found %d",i);
	 return i;
      };
   logprintf(LOG_DEBUG,'M',"find_spot: no spot found");
   return -1;
};

/* th is where to make the fog, who is who is being teleported */
static void spawn_fog(LevData *ld,int th,int who,int prid) {
   const ThingDyn *td=ldthingd(ld)+th;
   const ThingDyn *tdwho=ldthingd(ld)+who;
   fixed x=td->x,y=td->y,z=td->z;
   x+=(9*fixcos(td->angle))/8;
   y+=(9*fixsin(td->angle))/8;
   z+=(3*tdwho->proto->height)/4;
   new_thing(ld,prid,x,y,z);
};

static void perform(LevData *ld,const LT_Action *lta,int on,int targ) {
   MapEvent *me;
   /* deal with spawntype */
   if(lta->spawntype>0) {
      spawn_fog(ld,on,on,lta->spawntype);
      if(targ>=0) spawn_fog(ld,targ,on,lta->spawntype);
   };
   /* prevent queuing of events that are already happening */ 
   if((lta->delay==0)&&
      find_active_event(ld,lta->lumptype,lta->eventtype,on)) return;
   /* do we always want this to happen? */
   unqueue_event(ld,lta->lumptype,lta->eventtype,on);
   /* OK, create that event */
   me=insert_event(ld,lta->lumptype,lta->eventtype,on);
   /* deal with delay & sounds */
   me->start+=lta->delay;
   me->sound=lta->sound;
   me->stopsound=lta->stopsound;
   /* deal with targ */
   if(targ>=0) {
      ThingDyn *td=ldthingd(ld)+targ;
      me->x=td->x;
      me->y=td->y;
      me->z=td->z;
      me->angle=td->angle;
   };
};

void thing_hit_wall(LevData *ld,int thing,int wall,int isback,WallHitType t)  {
   int front,back,i;
   ThingDyn *td=ldthingd(ld)+thing;
   LineDyn *wd=ldlined(ld)+wall;
   const LineData *w=ldline(ld)+wall;
   const LineType *lt=lookup_linetype(w->type);
   if(lt==NULL) return;

   /* check reactivation */
   if(wd->block_activation) return;

   /* check direction */
   if((lt->flags&LT_FRONT_ONLY)&&isback) return;

   /* check to see whether this wall is being activated in appropriate way */
   switch(t) {
   case(Crossed): if(lt->flags&LT_ON_CROSSED) break; else return;
   case(Thumped): if(lt->flags&LT_ON_THUMPED) break; else return;
   case(Activated): if(lt->flags&LT_ON_ACTIVATED) break; else return;
   case(Damaged): if(lt->flags&LT_ON_DAMAGED) break; else return;
   default: 
      logprintf(LOG_ERROR,'M',
		"strange WallHitType (%d) in thing_hit_wall()",t);
      return;
   };

   /* check if player / monster is allowed */
   if((td->proto->flags&PT_PLAYER)&&!(lt->flags&LT_ALLOW_PLAYER)) {
      logprintf(LOG_DEBUG,'M',"player is not allowed to activate type %d",
		(int)(ldline(ld)[wall].type));
      return;
   };
   if(!(td->proto->flags&PT_PLAYER)&&!(lt->flags&LT_ALLOW_NONPLAYER)) {
      logprintf(LOG_DEBUG,'M',"non-player is not allowed to activate type %d",
		(int)(ldline(ld)[wall].type));
      return;
   };

   /* check keytype */
   if(lt->keytype&&!gettable_chk_key(lt->keytype)) {
      if(!wd->had_keymsg) game_message("YOU NEED A KEY!");
      wd->had_keymsg=1;
      return;
   };

   /* looks like we can go ahead */
   /*logprintf(LOG_DEBUG,'M',
             "line %d is having something done to it by thing %d"
	     ,wall,thing);*/

   /* block future activation if effect is not repeatable */
   if(!(lt->flags&LT_REPEATABLE)) wd->block_activation=1;

   /* make sound? */

   /* find front + back sides of wall for manual ops */
   front=ldline(ld)[wall].side[0];
   back=ldline(ld)[wall].side[1];

   /* perform actions */ 
   for(i=0;i<MAX_LT_ACTIONS;i++) { 
      const LT_Action *lta=lt->action+i; 
      if(NO_ACTION(lta)) break;
      switch(lta->lumptype) { 
	 
      case(ML_SECTOR):
	 /* deal with "manual" actions */ 
	 if(lta->flags&LTA_MANUAL) {
	    if(back<0||ldside(ld)[back].sector<0) 
	       logprintf(LOG_ERROR,'M',
			 "tried manual action on wall with no back?");
	    else
	       perform(ld,lta,ldside(ld)[back].sector,-1); 
	 } 
	 else if(lta->flags&LTA_MANUAL_FRONT) {
	    if(front<0||ldside(ld)[front].sector<0) 
	       logprintf(LOG_ERROR,'M',
			 "tried manual_f action on wall with no front?");
	    else
	       perform(ld,lta,ldside(ld)[front].sector,-1); 
	 } 
	 /* act on all tagged sectors */ 
	 else { 
	    int j,tag=ldline(ld)[wall].tag;
	    for(j=0;j<ldnsectors(ld);j++) 
	       if(ldsector(ld)[j].tag==tag) 
		  perform(ld,lta,j,-1);
	 };
	 break;

      case(ML_SIDE):
	 if(lta->flags&LTA_MANUAL) perform(ld,lta,back,-1); 
	 if(lta->flags&LTA_MANUAL_FRONT) perform(ld,lta,front,-1); 
	 break;

      case(ML_THING):
	 if(lt->spottype) perform(ld,lta,thing,find_spot(ld,lt,w->tag));
	 else perform(ld,lta,thing,-1); 
	 break;

      case(ML_LINE):
	 perform(ld,lta,wall,-1); 
	 break;

	 /* linetype refered to an action on some strange lumptype */
      default: 
	 logprintf(LOG_ERROR,'M',
		   "strange lumptype (%d) while performing actions",
		   (int)(lta->lumptype)); 
	 break; 
      }; 
   };
};












